/**
 * sendmail mqueue viewer v1.0
 * http://www.oook.cz/bsd/mqv/
 * Copyright (c) 2007 Jan Hornyak <pav@oook.cz>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <curses.h>
#include <glib.h>

#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>

typedef struct _QueueEntry QueueEntry;

struct _QueueEntry {
	gchar* path;
	gchar* reason;
	time_t filed;
	time_t last_change;
	gchar* sender;
	gchar* recipient;
	gchar* subject;
	gboolean valid;
};

struct MaxLengths {
	gint subject;
	gint sender;
	gint recipient;
	gint reason;
};

static void delete_mail();
static void finish(int signal);
static void invalidate_entry(GQuark id, gpointer data, gpointer user_data);
static int lookup_cursor(GQuark id);
static void print_body();
static void print_queue(int i);
static void remove_invalid_entry(GQuark id, gpointer data, gpointer user_data);
static void scan_queue(gchar* path);
static void scan_queues();

#define MQUEUE_DIR "/var/spool/mqueue"
#define CLIENT_MQUEUE_DIR "/var/spool/clientmqueue"

GData *queue;
GArray *queuear;
struct MaxLengths maxlengths;
GQuark cursor = 0;
int offset = 0;

#define MAXL(str) if (strlen(qe->str) > maxlengths.str || maxlengths.str == 0) maxlengths.str = strlen(qe->str);
#define last_row (slice_end - 1)

int main(int argc, char **argv) {

	int i, x, y, lines_on_screen, slice_end;

	/* set up signal handlers */
	signal(SIGINT, finish);
	signal(SIGTERM, finish);

	/* init ncurses */
	initscr();
	nonl();
	cbreak();
	keypad(stdscr, TRUE);
	noecho();

	start_color();
	for (i = 0; i < 8; i++) init_pair(i, i, COLOR_BLACK);

	refresh();

	getmaxyx(stdscr, y, x);
	lines_on_screen = (y - 1) / 2;

	/* create data list */
	g_datalist_init(&queue);
	queuear = g_array_new(FALSE, FALSE, sizeof(GQuark));

	scan_queues();

	/* loop */
	for (;;) {

		erase();

		/* set cursor if off bounds */
		if (cursor == 0 && queuear->len)
			cursor = g_array_index(queuear, GQuark, 0);

		/* determine visible slice of array */
		if (lines_on_screen < queuear->len) {
			slice_end = offset + lines_on_screen;
			if (queuear->len < slice_end) slice_end = queuear->len;
		} else {
			slice_end = queuear->len;
		}

		/* redraw the list */
		for (i = offset; i < slice_end; i++) {
			print_queue(i);
		}

		/* status line */
		mvprintw(y-1, 0, "%d mail%s", queuear->len, (queuear->len > 1 ? "s" : ""));
		if (lines_on_screen < queuear->len) {
			printw(" (%d-%d displayed)", offset+1, slice_end);
		}
		printw("   [r] refresh, [v] view, [d] delete");

		/* keyboard */
		int c = getch();

		switch (c) {
			case 'd':
				/* delete */
				if (cursor)
					delete_mail();
				break;
			case 'q':
				/* quit */
				finish(0);
				break;
			case 'r':
				/* refresh */
				printw("  -- refreshing --");
				refresh();
				scan_queues();
				break;
			case 'v':
			case 13: // ENTER
				/* display one mail detail */
				if (cursor) {
					erase();
					i = lookup_cursor(cursor);
					int saved_offset = offset;
					offset = i;
					print_queue(i);
					print_body();
					offset = saved_offset;
				}
				break;
			case KEY_UP:
				if (cursor) {
					i = lookup_cursor(cursor);
					if (i) {
						cursor = g_array_index(queuear, GQuark, i - 1);
						if (i <= offset) offset--;
					}
				}
				break;
			case KEY_PPAGE:
				if (cursor) {
					i = lookup_cursor(cursor);
					if (i) {
						if (i == offset) {
							if (i > lines_on_screen) {
								i -= lines_on_screen;
								offset -= lines_on_screen;
							} else {
								i = 0;
								offset = 0;
							}
						} else {
							i = offset;
						}
						cursor = g_array_index(queuear, GQuark, i);
					}
				}
				break;
			case KEY_DOWN:
				if (cursor) {
					i = lookup_cursor(cursor) + 1;
					if (i < queuear->len) {
						cursor = g_array_index(queuear, GQuark, i);
						if (i > last_row) offset++;
					}
				}
				break;
			case KEY_NPAGE:
				if (cursor) {
					i = lookup_cursor(cursor);
					if (i < queuear->len - 1) {
						if (i == last_row) {
							i += lines_on_screen;
							if (i >= queuear->len) i = queuear->len - 1;
							offset += lines_on_screen;
						} else {
							i = last_row;
						}
						cursor = g_array_index(queuear, GQuark, i);
					}
				}
				break;
		}
	}

	finish(0);
}

static void delete_mail() {

	QueueEntry *qe = g_datalist_id_get_data(&queue, cursor);
	const gchar *messid = g_quark_to_string(cursor);

	// unlink qf-filename
	gchar *basename = g_strconcat("qf", messid, NULL);
	gchar *filename = g_build_filename(qe->path, basename, NULL);
	unlink(filename);
	g_free(basename);
	g_free(filename);

	// unlink df-filename
	basename = g_strconcat("df", messid, NULL);
	filename = g_build_filename(qe->path, basename, NULL);
	unlink(filename);
	g_free(basename);
	g_free(filename);

	// refresh
	scan_queues();
}

static void finish(int signal) {

	g_datalist_clear(&queue);
	g_array_free(queuear, TRUE);
	endwin();

	exit(0);
}

static void invalidate_entry(GQuark id, gpointer data, gpointer user_data) {
	QueueEntry *qe = data;
	qe->valid = FALSE;
}

static int lookup_cursor(GQuark id) {
	int i, index = -1;
	for (i = 0; i < queuear->len; i++) {
		if (g_array_index(queuear, GQuark, i) == id) index = i;
	}
	g_assert(index >= 0);
	return index;
}

static void print_body() {

	// construct filename
	QueueEntry *qe = g_datalist_id_get_data(&queue, cursor);
	const gchar *messid = g_quark_to_string(cursor);
	gchar *basename = g_strconcat("df", messid, NULL);
	gchar *filename = g_build_filename(qe->path, basename, NULL);
	g_free(basename);

	// open file
	GError *err = NULL;
	GIOChannel *ch = g_io_channel_new_file(filename, "r", &err);
	if (!ch) {
		g_error("Failure to read from %s: %s", filename, err->message);
		g_error_free(err);
	}
	g_io_channel_set_encoding(ch, NULL, NULL);

	// read contents
	int x = 0, y = 2, maxx, maxy;
	getmaxyx(stdscr, maxy, maxx);
	move(y, x);
	gchar* line;
	while (g_io_channel_read_line(ch, &line, NULL, NULL, &err) == G_IO_STATUS_NORMAL) {
		printw("%s", line);
		g_free(line);
		getyx(stdscr, y, x);
		if (y == maxy - 1) {
			printw("-- more -- or [q]uit");
			int c = getch();
			if (c == 'q') goto out;
			erase();
			print_queue(offset);
			move(2, 0);
		}
	}
	getch();
out:
	if (err) g_error("Failure reading from %s: %s", filename, err->message);
	g_io_channel_close(ch);
	g_free(filename);

}

static void print_queue(int i) {
	GQuark id = g_array_index(queuear, GQuark, i);
	g_assert(id);
	QueueEntry *qe = g_datalist_id_get_data(&queue, id);
	int lineno = (i - offset) * 2;
	g_assert(qe);
	gchar *filed = g_strdup(ctime(&qe->filed));
	// gchar *last = g_strdup(ctime(&qe->last_change)); g_strchomp(last),
	mvprintw(lineno, 2, "%s  %s", g_quark_to_string(id), qe->sender);
	mvprintw(lineno, 20 + MAX(maxlengths.sender,maxlengths.recipient), "%s", qe->subject);
	mvprintw(lineno, 22 + MAX(maxlengths.sender,maxlengths.recipient) + MAX(maxlengths.subject,maxlengths.reason), "%s", g_strchomp(filed));
	mvprintw(lineno + 1, 18, "%s", qe->recipient);
	if (qe->reason) mvprintw(lineno + 1, 20 + MAX(maxlengths.sender,maxlengths.recipient), "%s", qe->reason);
	g_free(filed);
	// g_free(last);
	// mvprintw(lineno + 1, 0, "%d", id);
	if (id == cursor) {
		mvprintw(lineno, 0, ">");
		mvchgat(lineno, 0, -1, A_BOLD, COLOR_WHITE, NULL);
		mvchgat(lineno + 1, 0, -1, A_BOLD, COLOR_WHITE, NULL);
	} else {
		mvchgat(lineno, 0, -1, A_NORMAL, COLOR_WHITE, NULL);
		mvchgat(lineno + 1, 0, -1, A_NORMAL, COLOR_WHITE, NULL);
	}
}

static void remove_invalid_entry(GQuark id, gpointer data, gpointer user_data) {
	QueueEntry *qe = data;
	if (qe->valid != TRUE) {
		if (cursor == id) {
			if (lookup_cursor(cursor) + 1 == queuear->len) {
				if (queuear->len == 1)
					cursor = 0;
				else
					cursor = g_array_index(queuear, GQuark, queuear->len - 2);
			} else
				cursor = g_array_index(queuear, GQuark, lookup_cursor(cursor) + 1);

		}
		g_datalist_id_remove_data(&queue, id);
		g_array_remove_index(queuear, lookup_cursor(id));
	}
}

static void scan_queues() {

	maxlengths.subject = 0;
	maxlengths.sender = 0;
	maxlengths.recipient = 0;
	maxlengths.reason = 0;

	/* invalidate entries */
	g_datalist_foreach(&queue, invalidate_entry, NULL);

	scan_queue(MQUEUE_DIR);
	scan_queue(CLIENT_MQUEUE_DIR);

	/* remove invalid entries */
	g_datalist_foreach(&queue, remove_invalid_entry, NULL);

}

static void scan_queue(gchar *path) {

	GDir* mqueue_dir;
	const gchar* filename;

	GError *err = NULL;
	mqueue_dir = g_dir_open(path, 0, &err);
	if (!mqueue_dir) g_error("Cannot open %s: %s", path, err->message);

	while ((filename = g_dir_read_name(mqueue_dir))) {

		// skip non-q files
		if (strlen(filename) < 3) continue;
		if (!g_str_has_prefix(filename, "qf")) continue;

		// open q-file
		GIOChannel *ch;
		gchar *qfilename = g_build_filename(path, filename, NULL);
		ch = g_io_channel_new_file(qfilename, "r", &err);
		if (!ch) {
			g_error("Failure to read from %s: %s", qfilename, err->message);
			g_error_free(err);
		}
		g_io_channel_set_encoding(ch, NULL, NULL);

		// check if this is a new entry
		GQuark id = g_quark_from_string(filename+2);
		QueueEntry* qe = g_datalist_id_get_data(&queue, id);
		if (qe == NULL) {
			// create new data list entry
			qe = g_new0(QueueEntry, 1);
			g_array_append_val(queuear, id);
			g_datalist_id_set_data(&queue, id, qe);
		}
		qe->path = g_strdup(path);
		qe->valid = TRUE;

		// read contents
		gchar* line;
		gsize* length = 0;
		while ((g_io_channel_read_line(ch, &line, length, NULL, &err) == G_IO_STATUS_NORMAL)) {
			switch (line[0]) {
				case 'H':
					if (g_str_has_prefix(line,"H??Subject: ")) {
						qe->subject = g_strdup(g_strchomp(line+12));
						MAXL(subject);
					}
					break;
				case 'K':
					qe->last_change = g_ascii_strtoull(line+1, NULL, 10);
					break;
				case 'M':
					qe->reason = g_strdup(g_strchomp(line+1));
					MAXL(reason);
					break;
				case 'R':
					if (g_str_has_prefix(line,"RPFD:")) qe->recipient = g_strdup(g_strchomp(line+5));
					else qe->recipient = g_strdup(g_strchomp(line+1));
					MAXL(recipient);
					break;
				case 'S':
					qe->sender = g_strdup(g_strchomp(line+1));
					MAXL(sender);
					break;
				case 'T':
					qe->filed = g_ascii_strtoull(line+1, NULL, 10);
					break;
			}
			g_free(length);
			g_free(line);
		}
		if (err) g_error("Failure reading from %s: %s", qfilename, err->message);
		g_io_channel_close(ch);
		g_free(qfilename);
	}
	g_dir_close(mqueue_dir);
}
